@@ -22,18 +22,27 @@ from sphinx import addnodes
 from sphinx.roles import XRefRole
 from sphinx.domains import Domain, ObjType, Index
 from sphinx.directives import ObjectDescription, directives
+from sphinx.util import logging
 from sphinx.util.nodes import make_refnode
 from sphinx.util.docfields import GroupedField, TypedField
+from sphinx.util.docutils import Reporter, LoggingReporter
+from sphinx.locale import _
 
-# The env.get_doctree() lookup results in a pickle.load() call which is
-# expensive enough to dominate the runtime entirely when the number of endpoints
-# and references is large enough. The doctrees are generated during the read-
-# phase and we can cache their lookup during the write-phase significantly
-# improving performance.
-# Currently sphinxcontrib-httpdomain does not declare to support parallel read
-# support (parallel_read_safe is the default False) so we can simply use a
-# module global to hold the cache.
-_doctree_cache = {}
+logger = logging.getLogger(__name__)
+
+
+class DummyDocument(object):
+    """Used where the signature requires a docutils.node.Document but only its reporter
+    is being used.
+
+    Up until the current Sphinx 2.3.1 calls to env.get_doctree() (which would get
+    us said docutils.node.Document) result in pickle.load() calls which are expensive
+    enough to dominate the runtime entirely when the number of endpoints and references
+    is large enough.
+    """
+
+    def __init__(self, reporter):
+        self.reporter = reporter
 
 
 class DocRef(object):
@@ -55,7 +64,7 @@ class RFC2616Ref(DocRef):
     In 2014, RFC2616 was replaced by multiple RFCs (7230-7237)."""
 
     def __init__(self, section):
-        url = 'http://www.w3.org/Protocols/rfc2616/rfc2616-sec{0:d}.html'
+        url = 'https://www.w3.org/Protocols/rfc2616/rfc2616-sec{0:d}.html'
         url = url.format(int(section))
         super(RFC2616Ref, self).__init__(url, 'sec', section)
 
@@ -64,14 +73,14 @@ class IETFRef(DocRef):
     """Represents a reference to the specific IETF RFC."""
 
     def __init__(self, rfc, section):
-        url = 'http://tools.ietf.org/html/rfc{0:d}'.format(rfc)
+        url = 'https://tools.ietf.org/html/rfc{0:d}'.format(rfc)
         super(IETFRef, self).__init__(url, 'section-', section)
 
 
 class EventSourceRef(DocRef):
 
     def __init__(self, section):
-        url = 'http://www.w3.org/TR/eventsource/'
+        url = 'https://www.w3.org/TR/eventsource/'
         super(EventSourceRef, self).__init__(url, section, '')
 
 
@@ -79,7 +88,7 @@ class CORSRef(DocRef):
     """Represents a reference to W3 Cross-Origin Resource Sharing recommendation."""
 
     def __init__(self, name, type):
-        url = 'http://www.w3.org/TR/cors/'
+        url = 'https://www.w3.org/TR/cors/'
         super(CORSRef, self).__init__(url, name, '-' + type)
 
 
@@ -124,7 +133,7 @@ HEADER_REFS = {
     'Cookie': IETFRef(2109, '4.3.4'), # also RFC6265 section 5.4
     'Date': IETFRef(7231, '7.1.1.2'),
     'Destination': IETFRef(2518, 9.3),
-    'ETag': IETFRef(7231, 2.3),
+    'ETag': IETFRef(7232, 2.3),
     'Expect': IETFRef(7231, '5.1.1'),
     'Expires': IETFRef(7234, 5.3),
     'From': IETFRef(7231, '5.5.2'),
@@ -259,27 +268,24 @@ class HTTPResource(ObjectDescription):
     doc_field_types = [
         TypedField('parameter', label='Parameters',
                    names=('param', 'parameter', 'arg', 'argument'),
-                   typerolename='obj', typenames=('paramtype', 'type')),
+                   typenames=('paramtype', 'type')),
         TypedField('jsonparameter', label='JSON Parameters',
                    names=('jsonparameter', 'jsonparam', 'json'),
-                   typerolename='obj', typenames=('jsonparamtype', 'jsontype')),
+                   typenames=('jsonparamtype', 'jsontype')),
         TypedField('requestjsonobject', label='Request JSON Object',
                    names=('reqjsonobj', 'reqjson', '<jsonobj', '<json'),
-                   typerolename='obj', typenames=('reqjsonobj', '<jsonobj')),
+                   typenames=('reqjsonobj', '<jsonobj')),
         TypedField('requestjsonarray', label='Request JSON Array of Objects',
                    names=('reqjsonarr', '<jsonarr'),
-                   typerolename='obj',
                    typenames=('reqjsonarrtype', '<jsonarrtype')),
         TypedField('responsejsonobject', label='Response JSON Object',
                    names=('resjsonobj', 'resjson', '>jsonobj', '>json'),
-                   typerolename='obj', typenames=('resjsonobj', '>jsonobj')),
+                   typenames=('resjsonobj', '>jsonobj')),
         TypedField('responsejsonarray', label='Response JSON Array of Objects',
                    names=('resjsonarr', '>jsonarr'),
-                   typerolename='obj',
                    typenames=('resjsonarrtype', '>jsonarrtype')),
         TypedField('queryparameter', label='Query Parameters',
                    names=('queryparameter', 'queryparam', 'qparam', 'query'),
-                   typerolename='obj',
                    typenames=('queryparamtype', 'querytype', 'qtype')),
         GroupedField('formparameter', label='Form Parameters',
                      names=('formparameter', 'formparam', 'fparam', 'form')),
@@ -410,6 +416,9 @@ class HTTPXRefRole(XRefRole):
             title = self.method.upper() + ' ' + title
         return title, target
 
+    def result_nodes(self, document, env, node, is_ref):
+        return [node], []
+
 
 class HTTPXRefMethodRole(XRefRole):
 
@@ -485,17 +494,19 @@ class HTTPXRefStatusRole(XRefRole):
         elif status is None:
             return report_unknown_code()
         elif code == 226:
-            url = 'http://www.ietf.org/rfc/rfc3229.txt'
+            url = 'https://www.ietf.org/rfc/rfc3229.txt'
         elif code == 418:
-            url = 'http://www.ietf.org/rfc/rfc2324.txt'
+            url = 'https://www.ietf.org/rfc/rfc2324.txt'
         elif code == 429:
-            url = 'http://tools.ietf.org/html/rfc6585#section-4'
+            url = 'https://tools.ietf.org/html/rfc6585#section-4'
         elif code == 449:
-            url = 'http://msdn.microsoft.com/en-us/library/dd891478(v=prot.10).aspx'
+            url = 'https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wdv/83ecf19f-e0f8-4706-aae5-ba618f52f100'
+        elif code == 451:
+            url = 'https://www.ietf.org/rfc/rfc7725.txt'
         elif code in WEBDAV_STATUS_CODES:
-            url = 'http://tools.ietf.org/html/rfc4918#section-11.%d' % (WEBDAV_STATUS_CODES.index(code) + 1)
+            url = 'https://tools.ietf.org/html/rfc4918#section-11.%d' % (WEBDAV_STATUS_CODES.index(code) + 1)
         elif code in HTTP_STATUS_CODES:
-            url = 'http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html' \
+            url = 'https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html' \
                   '#sec10.' + ('%d.%d' % (code // 100, 1 + code % 100))
         else:
             url = ''
@@ -571,17 +582,17 @@ class HTTPDomain(Domain):
     label = 'HTTP'
 
     object_types = {
-        'options': ObjType('options', 'options', 'obj'),
-        'head': ObjType('head', 'head', 'obj'),
-        'post': ObjType('post', 'post', 'obj'),
-        'get': ObjType('get', 'get', 'obj'),
-        'put': ObjType('put', 'put', 'obj'),
-        'patch': ObjType('patch', 'patch', 'obj'),
-        'delete': ObjType('delete', 'delete', 'obj'),
-        'trace': ObjType('trace', 'trace', 'obj'),
-        'connect': ObjType('connect', 'connect', 'obj'),
-        'copy': ObjType('copy', 'copy', 'obj'),
-        'any': ObjType('any', 'any', 'obj')
+        'options': ObjType('options', 'options'),
+        'head': ObjType('head', 'head'),
+        'post': ObjType('post', 'post'),
+        'get': ObjType('get', 'get'),
+        'put': ObjType('put', 'put'),
+        'patch': ObjType('patch', 'patch'),
+        'delete': ObjType('delete', 'delete'),
+        'trace': ObjType('trace', 'trace'),
+        'connect': ObjType('connect', 'connect'),
+        'copy': ObjType('copy', 'copy'),
+        'any': ObjType('any', 'any')
     }
 
     directives = {
@@ -651,10 +662,10 @@ class HTTPDomain(Domain):
             if role is None:
                 return None
 
-            if fromdocname not in _doctree_cache:
-                _doctree_cache[fromdocname] = env.get_doctree(fromdocname)
-            doctree = _doctree_cache[fromdocname]
-
+            reporter = LoggingReporter(env.doc2path(fromdocname),
+                                       report_level=Reporter.WARNING_LEVEL,
+                                       halt_level=Reporter.SEVERE_LEVEL)
+            doctree = DummyDocument(reporter)
             resnode = role.result_nodes(doctree, env, node, None)[0][0]
             if isinstance(resnode, addnodes.pending_xref):
                 text = node[0][0]
@@ -689,6 +700,27 @@ class HTTPDomain(Domain):
                 anchor = http_resource_anchor(method, path)
                 yield (path, path, method, info[0], anchor, 1)
 
+    def merge_domaindata(self, docnames, otherdata):
+        """Merge domaindata from the workers/chunks when they return.
+
+        Called once per parallelization chunk.
+        Only used when sphinx is run in parallel mode.
+
+        :param docnames: a Set of the docnames that are part of the current chunk to merge
+        :param otherdata: the partial data calculated by the current chunk
+        """
+        for typ in self.object_types:
+            self_data = self.data[typ]
+            other_data = otherdata[typ]
+            for entry_point_name, entry_point_data in other_data.items():
+                if entry_point_name in self_data:
+                    logger.warning('duplicate HTTP %s method definition %s in %s, '
+                                   'other instance is in %s' %
+                                   (typ, entry_point_name,
+                                    self.env.doc2path(other_data[entry_point_name][0]),
+                                    self.env.doc2path(self_data[entry_point_name][0])))
+                else:
+                    self_data[entry_point_name] = entry_point_data
 
 class HTTPLexer(RegexLexer):
     """Lexer for HTTP sessions."""
@@ -755,8 +787,17 @@ class HTTPLexer(RegexLexer):
     }
 
 
+def register_routingtable_as_label(app, document):
+    labels = app.env.domaindata['std']['labels']
+    labels['routingtable'] = 'http-routingtable', '', _('HTTP Routing Table')
+    anonlabels = app.env.domaindata['std']['anonlabels']
+    anonlabels['routingtable'] = 'http-routingtable', ''
+
+
 def setup(app):
     app.add_domain(HTTPDomain)
+    app.connect('doctree-read', register_routingtable_as_label)
+
     try:
         get_lexer_by_name('http')
     except ClassNotFound:
@@ -766,3 +807,5 @@ def setup(app):
     app.add_config_value('http_index_localname', 'HTTP Routing Table', True)
     app.add_config_value('http_strict_mode', True, None)
     app.add_config_value('http_headers_ignore_prefixes', ['X-'], None)
+    return {"parallel_read_safe": True,
+            "parallel_write_safe": True}
